# -*- coding: utf-8 -*-
import sys
import threading

import pytest
import pytz

import pytz_deprecation_shim as pds
from pytz_deprecation_shim import helpers as pds_helpers

from ._common import UTC

get_timezone = pds._compat.get_timezone
get_fixed_timezone = pds._compat.get_fixed_offset_zone

SYS_MODULES_LOCK = threading.Lock()
HELPER_LOCK = threading.Lock()

ZONE_LIST = (
    "America/New_York",
    "Asia/Tokyo",
    "Australia/Sydney",
    "Africa/Abidjan",
    "America/Santiago",
    "Europe/London",
    "Europe/Dublin",
)

OFFSET_MINUTES_LIST = (0, -1439, 1439, 0, 60, -60, 11)


def _timezones(tz_constructor):
    return tuple(map(tz_constructor, ZONE_LIST))


def _fixed_offsets(tz_constructor):
    return tuple(map(tz_constructor, OFFSET_MINUTES_LIST))


NON_PYTZ_ZONES = (
    (UTC, pds.timezone("UTC"))
    + _timezones(pds.timezone)
    + _timezones(get_timezone)
    + _fixed_offsets(pds.fixed_offset_timezone)
    + _fixed_offsets(get_fixed_timezone)
)

PYTZ_ZONES = (
    (pytz.UTC, pytz.timezone("UTC"))
    + _timezones(pytz.timezone)
    + _fixed_offsets(pytz.FixedOffset)
)


@pytest.fixture(autouse=True)
def helper_lock():
    """Lock around anything using the helper module."""
    with HELPER_LOCK:
        yield


@pytest.fixture
def no_pytz():
    """Fixture to remove pytz from sys.modules for the duration of the test."""
    with SYS_MODULES_LOCK:
        pds._common._PYTZ_IMPORTED = False
        base_classes = pds_helpers._PYTZ_BASE_CLASSES
        pds_helpers._PYTZ_BASE_CLASSES = None
        pytz_modules = {}

        for modname in list(sys.modules):
            if modname.split(".", 1)[0] != "pytz":  # pragma: nocover
                continue

            pytz_modules[modname] = sys.modules.pop(modname)

        try:
            yield
        finally:
            sys.modules.update(pytz_modules)
            pds_helpers._PYTZ_BASE_CLASSES = base_classes


@pytest.mark.parametrize("tz", NON_PYTZ_ZONES)
def test_not_pytz_zones(tz):
    """Tests is_pytz_zone for non-pytz zones."""
    assert not pds_helpers.is_pytz_zone(tz)


@pytest.mark.parametrize("tz", NON_PYTZ_ZONES)
def test_not_pytz_no_pytz(tz, no_pytz):
    """Tests is_pytz_zone when pytz has not been imported."""
    assert not pds_helpers.is_pytz_zone(tz)

    # Ensure that the test didn't import `pytz`.
    assert "pytz" not in sys.modules


@pytest.mark.parametrize("tz", PYTZ_ZONES)
def test_pytz_zone(tz):
    assert pds_helpers.is_pytz_zone(tz)


@pytest.mark.parametrize("key", ("UTC",) + ZONE_LIST)
def test_pytz_zones_before_after(key, no_pytz):
    """Tests is_pytz_zone when pytz is imported after first use.

    We want to make sure that is_pytz_zone doesn't inappropriately cache the
    fact that pytz hasn't been imported, and just always return ``False``, even
    if pytz is imported after the first call to is_pytz_zone.
    """

    non_pytz_zone = pds.timezone(key)
    assert not pds_helpers.is_pytz_zone(non_pytz_zone)
    assert "pytz" not in sys.modules

    import pytz

    pytz_zone = pytz.timezone(key)
    assert pds_helpers.is_pytz_zone(pytz_zone)


@pytest.mark.parametrize("key", ZONE_LIST)
def test_upgrade_shim_timezone(key):
    shim_zone = pds.timezone(key)

    actual = pds_helpers.upgrade_tzinfo(shim_zone)
    expected = get_timezone(key)

    assert actual is expected


@pytest.mark.parametrize("key", ZONE_LIST)
def test_upgrade_pytz_timezone(key):
    pytz_zone = pytz.timezone(key)

    actual = pds_helpers.upgrade_tzinfo(pytz_zone)
    expected = get_timezone(key)

    assert actual is expected


@pytest.mark.parametrize("offset_minutes", OFFSET_MINUTES_LIST)
def test_upgrade_shim_fixed_offset(offset_minutes):
    shim_zone = pds.fixed_offset_timezone(offset_minutes)

    actual = pds_helpers.upgrade_tzinfo(shim_zone)
    expected = get_fixed_timezone(offset_minutes)

    # get_fixed_timezone doesn't do any caching, so the fixed offsets aren't
    # necessarily going to be singletons.
    assert actual == expected


@pytest.mark.parametrize("offset_minutes", OFFSET_MINUTES_LIST)
def test_upgrade_pytz_fixed_offset(offset_minutes):
    pytz_zone = pytz.FixedOffset(offset_minutes)

    actual = pds_helpers.upgrade_tzinfo(pytz_zone)
    expected = get_fixed_timezone(offset_minutes)

    # get_fixed_timezone doesn't do any caching, so the fixed offsets aren't
    # necessarily going to be singletons.
    assert actual == expected


@pytest.mark.parametrize("utc", (pds.UTC, pytz.UTC))
def test_upgrade_utc(utc):
    """Tests that upgrade_zone called on UTC objects returns _compat.UTC.

    This is relevant because tz.gettz("UTC") or zoneinfo.ZoneInfo("UTC") may
    each return a tzfile-based zone, not their respective UTC singletons.
    """
    actual = pds_helpers.upgrade_tzinfo(utc)

    assert actual is UTC


@pytest.mark.parametrize(
    "tz",
    [UTC]
    + list(map(get_timezone, ZONE_LIST))
    + list(map(get_fixed_timezone, OFFSET_MINUTES_LIST)),
)
def test_upgrade_tz_noop(tz):
    """Tests that non-shim, non-pytz zones are unaffected by upgrade_tzinfo."""
    actual = pds_helpers.upgrade_tzinfo(tz)

    assert actual is tz



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os


def read_version():
    here = os.path.split(__file__)[0]
    version_file = os.path.join(here, "../VERSION")

    with open(version_file, "rt") as f:
        return f.read().strip()


VERSION = read_version()


# -- Project information -----------------------------------------------------

project = "pytz_deprecation_shim"
copyright = "2020, Paul Ganssle"
author = "Paul Ganssle"

# The full version, including alpha/beta/rc tags
release = VERSION


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# For cross-links to other documentation
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.9", None),
    "dateutil": ("https://dateutil.readthedocs.io/en/stable/", None),
}

_repo = "https://github.com/pganssle/pytz-deprecation-shim/"
extlinks = {
    "gh": (_repo + "issues/%s", "GH-"),
    "gh-pr": (_repo + "pull/%s", "GH-"),
    "pypi": ("https://pypi.org/project/%s", ""),
    "bpo": ("https://bugs.python.org/issue%s", "bpo-"),
    "cpython-pr": ("https://github.com/python/cpython/pull/%s", "CPython PR #"),
}



from datetime import datetime, timedelta

import hypothesis
import pytest
import pytz

import pytz_deprecation_shim as pds

from ._common import (
    invalid_offset_minute_strategy,
    invalid_zone_strategy,
    valid_zone_strategy,
)


# On Python 2.7 when running 5 test suites in parallel, this is reliably flaky
# without increasing the deadline. When running 1 or 2 in parallel it's only
# intermittently flaky.
@hypothesis.settings(deadline=timedelta(seconds=30))
@hypothesis.given(key=invalid_zone_strategy)
@hypothesis.example(key="")
def test_invalid_zones(key):
    with pytest.raises(pds.UnknownTimeZoneError) as exc_info:
        pds.timezone(key)

    assert issubclass(exc_info.type, pytz.UnknownTimeZoneError)


@hypothesis.given(key=valid_zone_strategy)
def test_localize_aware_datetime(key):
    zone = pds.timezone(key)
    dt = datetime(2020, 1, 1, tzinfo=pds.UTC)

    with pytest.warns(pds.PytzUsageWarning), pytest.raises(ValueError):
        zone.localize(dt)


@hypothesis.given(key=valid_zone_strategy)
def test_normalize_naive_datetime(key):
    zone = pds.timezone(key)
    dt = datetime(2020, 1, 1)

    with pytest.warns(pds.PytzUsageWarning), pytest.raises(ValueError):
        zone.normalize(dt)


@hypothesis.given(offset=invalid_offset_minute_strategy)
def test_invalid_fixed_offsets(offset):
    with pytest.raises(ValueError):
        pds.fixed_offset_timezone(offset)



# -*- coding: utf-8 -*-
import base64
import io
import json
import os
import zlib
from datetime import datetime, timedelta

import attr

from ._common import (
    MAX_DATETIME,
    MIN_DATETIME,
    PY2,
    UTC,
    ZERO,
    get_fold,
    lru_cache,
    round_timedelta,
)

ONE_H = timedelta(hours=1)
ZONEINFO_DATA = None
ZONEINFO_DATA_VERSION = None


def _decode_tzinfo(data):
    raw_data = b"".join((s.encode() for s in data))
    decoded = base64.b64decode(raw_data)

    return zlib.decompress(decoded)


def _load_zoneinfo_json():
    global ZONEINFO_DATA
    global ZONEINFO_DATA_VERSION

    here = os.path.split(__file__)[0]
    data_file = os.path.join(here, "data", "zoneinfo_data.json")
    with open(data_file, "rt") as f:
        zi = json.load(f)

    ZONEINFO_DATA_VERSION = zi["metadata"]["version"]
    ZONEINFO_DATA = {
        key: _decode_tzinfo(data) for key, data in zi["data"].items()
    }


def get_zone_file_obj(key):
    if ZONEINFO_DATA is None:
        _load_zoneinfo_json()

    return io.BytesIO(ZONEINFO_DATA[key])


@lru_cache(None)
def get_unambiguous_cases():
    """Returns a list of unambiguous test cases for parameterization.

    The list has the form "key,dt,offset", where each value for transition
    is a ZoneTransition object.
    """
    cases = []

    def _add_case(key, dt, offset):
        if not MIN_DATETIME < dt < MAX_DATETIME:
            return

        cases.append((key, dt, offset))

    for key in ZoneDumpData.transition_keys():
        for transition in ZoneDumpData.load_transition_examples(key):
            _add_case(
                key,
                transition.transition - timedelta(days=2),
                transition.offset_before,
            )

            _add_case(
                key,
                transition.transition + timedelta(days=2),
                transition.offset_after,
            )

    for key, offset in ZoneDumpData.fixed_offset_zones():
        fixed_offset_cases = [
            (key, datetime(2000, 1, 1), offset),
            (key, datetime(2020, 2, 29), offset),
            (key, datetime(1900, 1, 1), offset),
            (key, datetime(2040, 1, 1), offset),
        ]

        for case in fixed_offset_cases:
            _add_case(*case)

    return tuple(cases)


@lru_cache(None)
def _get_folds_and_gaps():
    fold_cases = []
    gap_cases = []

    for key in ZoneDumpData.transition_keys():
        for transition in ZoneDumpData.load_transition_examples(key):
            if not MIN_DATETIME < transition.transition < MAX_DATETIME:
                continue

            if transition.fold:
                cases = fold_cases
            elif transition.gap:
                cases = gap_cases
            else:  # pragma: nocover
                continue

            cases.append([key, transition])

    return tuple(map(tuple, (fold_cases, gap_cases)))


@lru_cache(None)
def get_fold_cases():
    """Returns a list of test cases that involve an ambiguous time."""
    return _get_folds_and_gaps()[0]


@lru_cache(None)
def get_gap_cases():
    """Returns a list of test cases that involve an imaginary time."""
    return _get_folds_and_gaps()[1]


class ZoneDumpData(object):
    @classmethod
    def transition_keys(cls):
        return cls._get_zonedump().keys()

    @classmethod
    def load_transition_examples(cls, key):
        return cls._get_zonedump()[key]

    @classmethod
    def fixed_offset_zones(cls):
        if not cls._FIXED_OFFSET_ZONES:
            cls._populate_fixed_offsets()

        return cls._FIXED_OFFSET_ZONES.items()

    @classmethod
    def _get_zonedump(cls):
        if not cls._ZONEDUMP_DATA:
            cls._populate_zonedump_data()
        return cls._ZONEDUMP_DATA

    @classmethod
    def _populate_fixed_offsets(cls):
        cls._FIXED_OFFSET_ZONES = {
            "UTC": ZoneOffset("UTC", ZERO, ZERO),
        }

    @classmethod
    def _populate_zonedump_data(cls):
        def _Africa_Abidjan():
            LMT = ZoneOffset("LMT", timedelta(seconds=-968))
            GMT = ZoneOffset("GMT", ZERO)

            return [
                ZoneTransition(datetime(1912, 1, 1), LMT, GMT),
            ]

        def _Africa_Casablanca():
            P00_s = ZoneOffset("+00", ZERO, ZERO)
            P01_d = ZoneOffset("+01", ONE_H, ONE_H)
            P00_d = ZoneOffset("+00", ZERO, -ONE_H)
            P01_s = ZoneOffset("+01", ONE_H, ZERO)

            return [
                # Morocco sometimes pauses DST during Ramadan
                ZoneTransition(datetime(2018, 3, 25, 2), P00_s, P01_d),
                ZoneTransition(datetime(2018, 5, 13, 3), P01_d, P00_s),
                ZoneTransition(datetime(2018, 6, 17, 2), P00_s, P01_d),
                # On October 28th Morocco set standard time to +01,
                # with negative DST only during Ramadan
                ZoneTransition(datetime(2018, 10, 28, 3), P01_d, P01_s),
                ZoneTransition(datetime(2019, 5, 5, 3), P01_s, P00_d),
                ZoneTransition(datetime(2019, 6, 9, 2), P00_d, P01_s),
            ]

        def _America_Los_Angeles():
            LMT = ZoneOffset("LMT", timedelta(seconds=-28378), ZERO)
            PST = ZoneOffset("PST", timedelta(hours=-8), ZERO)
            PDT = ZoneOffset("PDT", timedelta(hours=-7), ONE_H)
            PWT = ZoneOffset("PWT", timedelta(hours=-7), ONE_H)
            PPT = ZoneOffset("PPT", timedelta(hours=-7), ONE_H)

            return [
                ZoneTransition(datetime(1883, 11, 18, 12, 7, 2), LMT, PST),
                ZoneTransition(datetime(1918, 3, 31, 2), PST, PDT),
                ZoneTransition(datetime(1918, 3, 31, 2), PST, PDT),
                ZoneTransition(datetime(1918, 10, 27, 2), PDT, PST),
                # Transition to Pacific War Time
                ZoneTransition(datetime(1942, 2, 9, 2), PST, PWT),
                # Transition from Pacific War Time to Pacific Peace Time
                ZoneTransition(datetime(1945, 8, 14, 16), PWT, PPT),
                ZoneTransition(datetime(1945, 9, 30, 2), PPT, PST),
                ZoneTransition(datetime(2015, 3, 8, 2), PST, PDT),
                ZoneTransition(datetime(2015, 11, 1, 2), PDT, PST),
                # After 2038: Rules continue indefinitely
                ZoneTransition(datetime(2450, 3, 13, 2), PST, PDT),
                ZoneTransition(datetime(2450, 11, 6, 2), PDT, PST),
            ]

        def _America_Santiago():
            LMT = ZoneOffset("LMT", timedelta(seconds=-16966), ZERO)
            SMT = ZoneOffset("SMT", timedelta(seconds=-16966), ZERO)
            N05 = ZoneOffset("-05", timedelta(seconds=-18000), ZERO)
            N04 = ZoneOffset("-04", timedelta(seconds=-14400), ZERO)
            N03 = ZoneOffset("-03", timedelta(seconds=-10800), ONE_H)

            return [
                ZoneTransition(datetime(1890, 1, 1), LMT, SMT),
                ZoneTransition(datetime(1910, 1, 10), SMT, N05),
                ZoneTransition(datetime(1916, 7, 1), N05, SMT),
                ZoneTransition(datetime(2008, 3, 30), N03, N04),
                ZoneTransition(datetime(2008, 10, 12), N04, N03),
                ZoneTransition(datetime(2040, 4, 8), N03, N04),
                ZoneTransition(datetime(2040, 9, 2), N04, N03),
            ]

        def _Asia_Tokyo():
            JST = ZoneOffset("JST", timedelta(seconds=32400), ZERO)
            JDT = ZoneOffset("JDT", timedelta(seconds=36000), ONE_H)

            # Japan had DST from 1948 to 1951, and it was unusual in that
            # the transition from DST to STD occurred at 25:00, and is
            # denominated as such in the time zone database
            return [
                ZoneTransition(datetime(1948, 5, 2), JST, JDT),
                ZoneTransition(datetime(1948, 9, 12, 1), JDT, JST),
                ZoneTransition(datetime(1951, 9, 9, 1), JDT, JST),
            ]

        def _Australia_Sydney():
            LMT = ZoneOffset("LMT", timedelta(seconds=36292), ZERO)
            AEST = ZoneOffset("AEST", timedelta(seconds=36000), ZERO)
            AEDT = ZoneOffset("AEDT", timedelta(seconds=39600), ONE_H)

            return [
                ZoneTransition(datetime(1895, 2, 1), LMT, AEST),
                ZoneTransition(datetime(1917, 1, 1, 0, 1), AEST, AEDT),
                ZoneTransition(datetime(1917, 3, 25, 2), AEDT, AEST),
                ZoneTransition(datetime(2012, 4, 1, 3), AEDT, AEST),
                ZoneTransition(datetime(2012, 10, 7, 2), AEST, AEDT),
                ZoneTransition(datetime(2040, 4, 1, 3), AEDT, AEST),
                ZoneTransition(datetime(2040, 10, 7, 2), AEST, AEDT),
            ]

        def _Europe_Dublin():
            LMT = ZoneOffset("LMT", timedelta(seconds=-1500), ZERO)
            DMT = ZoneOffset("DMT", timedelta(seconds=-1521), ZERO)
            IST_0 = ZoneOffset("IST", timedelta(seconds=2079), ONE_H)
            GMT_0 = ZoneOffset("GMT", ZERO, ZERO)
            BST = ZoneOffset("BST", ONE_H, ONE_H)
            GMT_1 = ZoneOffset("GMT", ZERO, -ONE_H)
            IST_1 = ZoneOffset("IST", ONE_H, ZERO)

            return [
                ZoneTransition(datetime(1880, 8, 2, 0), LMT, DMT),
                ZoneTransition(datetime(1916, 5, 21, 2), DMT, IST_0),
                ZoneTransition(datetime(1916, 10, 1, 3), IST_0, GMT_0),
                ZoneTransition(datetime(1917, 4, 8, 2), GMT_0, BST),
                ZoneTransition(datetime(2016, 3, 27, 1), GMT_1, IST_1),
                ZoneTransition(datetime(2016, 10, 30, 2), IST_1, GMT_1),
                ZoneTransition(datetime(2487, 3, 30, 1), GMT_1, IST_1),
                ZoneTransition(datetime(2487, 10, 26, 2), IST_1, GMT_1),
            ]

        def _Europe_Lisbon():
            WET = ZoneOffset("WET", ZERO, ZERO)
            WEST = ZoneOffset("WEST", ONE_H, ONE_H)
            CET = ZoneOffset("CET", ONE_H, ZERO)
            CEST = ZoneOffset("CEST", timedelta(seconds=7200), ONE_H)

            return [
                ZoneTransition(datetime(1992, 3, 29, 1), WET, WEST),
                ZoneTransition(datetime(1992, 9, 27, 2), WEST, CET),
                ZoneTransition(datetime(1993, 3, 28, 2), CET, CEST),
                ZoneTransition(datetime(1993, 9, 26, 3), CEST, CET),
                ZoneTransition(datetime(1996, 3, 31, 2), CET, WEST),
                ZoneTransition(datetime(1996, 10, 27, 2), WEST, WET),
            ]

        def _Europe_London():
            LMT = ZoneOffset("LMT", timedelta(seconds=-75), ZERO)
            GMT = ZoneOffset("GMT", ZERO, ZERO)
            BST = ZoneOffset("BST", ONE_H, ONE_H)

            return [
                ZoneTransition(datetime(1847, 12, 1), LMT, GMT),
                ZoneTransition(datetime(2005, 3, 27, 1), GMT, BST),
                ZoneTransition(datetime(2005, 10, 30, 2), BST, GMT),
                ZoneTransition(datetime(2043, 3, 29, 1), GMT, BST),
                ZoneTransition(datetime(2043, 10, 25, 2), BST, GMT),
            ]

        def _Pacific_Kiritimati():
            LMT = ZoneOffset("LMT", timedelta(seconds=-37760), ZERO)
            N1040 = ZoneOffset("-1040", timedelta(seconds=-38400), ZERO)
            N10 = ZoneOffset("-10", timedelta(seconds=-36000), ZERO)
            P14 = ZoneOffset("+14", timedelta(seconds=50400), ZERO)

            # This is literally every transition in Christmas Island history
            return [
                ZoneTransition(datetime(1901, 1, 1), LMT, N1040),
                ZoneTransition(datetime(1979, 10, 1), N1040, N10),
                # They skipped December 31, 1994
                ZoneTransition(datetime(1994, 12, 31), N10, P14),
            ]

        cls._ZONEDUMP_DATA = {
            "Africa/Abidjan": _Africa_Abidjan(),
            "Africa/Casablanca": _Africa_Casablanca(),
            "America/Los_Angeles": _America_Los_Angeles(),
            "America/Santiago": _America_Santiago(),
            "Australia/Sydney": _Australia_Sydney(),
            "Asia/Tokyo": _Asia_Tokyo(),
            "Europe/Dublin": _Europe_Dublin(),
            "Europe/Lisbon": _Europe_Lisbon(),
            "Europe/London": _Europe_London(),
            "Pacific/Kiritimati": _Pacific_Kiritimati(),
        }

    _ZONEDUMP_DATA = None
    _FIXED_OFFSET_ZONES = None


if PY2:
    td_converter = round_timedelta
else:
    td_converter = lambda x: x


@attr.s
class ZoneOffset(object):
    tzname = attr.ib(type=str)
    utcoffset = attr.ib(type=timedelta, converter=td_converter)
    dst = attr.ib(type=timedelta, default=ZERO, converter=td_converter)


@attr.s
class ZoneTransition(object):
    transition = attr.ib(type=datetime)
    offset_before = attr.ib(type=ZoneOffset)
    offset_after = attr.ib(type=ZoneOffset)

    @property
    def transition_utc(self):
        return (self.transition - self.offset_before.utcoffset).replace(
            tzinfo=UTC
        )

    @property
    def fold(self):
        """Whether this introduces a fold"""
        return self.offset_before.utcoffset > self.offset_after.utcoffset

    @property
    def gap(self):
        """Whether this introduces a gap"""
        return self.offset_before.utcoffset < self.offset_after.utcoffset

    @property
    def delta(self):
        return self.offset_after.utcoffset - self.offset_before.utcoffset

    @property
    def anomaly_start(self):
        if get_fold(self):
            return self.transition + self.delta
        else:
            return self.transition

    @property
    def anomaly_end(self):
        if not get_fold(self):
            return self.transition + self.delta
        else:
            return self.transition



import copy
import pickle
from datetime import datetime, timedelta, tzinfo

import hypothesis
import hypothesis.strategies as hst
import pytest
import pytz

import pytz_deprecation_shim as pds

from . import _zoneinfo_data
from ._common import (
    MAX_DATETIME,
    MIN_DATETIME,
    PY2,
    assert_dt_equivalent,
    assert_dt_offset,
    conditional_examples,
    datetime_unambiguous,
    dt_strategy,
    enfold,
    get_fold,
    offset_minute_strategy,
    round_normalized,
    valid_zone_strategy,
)

ONE_SECOND = timedelta(seconds=1)
ARBITRARY_KEY_STRATEGY = hst.from_regex("[a-zA-Z][a-zA-Z_]+(/[a-zA-Z_]+)+")


def _make_no_cache_timezone():
    if PY2:
        import dateutil.tz

        def no_cache_timezone(key, tz=dateutil.tz):
            return pds.wrap_zone(tz.gettz.nocache(key), key)

    else:
        try:
            import zoneinfo
        except ImportError:
            from backports import zoneinfo

        def no_cache_timezone(key, zoneinfo=zoneinfo):
            return pds.wrap_zone(zoneinfo.ZoneInfo.no_cache(key), key)

    return no_cache_timezone


no_cache_timezone = _make_no_cache_timezone()
del _make_no_cache_timezone


def test_fixed_offset_utc():
    """Tests that fixed_offset_timezone(0) always returns UTC."""
    assert pds.fixed_offset_timezone(0) is pds.UTC


@pytest.mark.parametrize("key", ["utc", "UTC"])
def test_timezone_utc_singleton(key):
    assert pds.timezone(key) is pds.UTC


@hypothesis.given(minutes=offset_minute_strategy.filter(lambda m: m != 0))
def test_str_fixed_offset(minutes):
    shim_zone = pds.fixed_offset_timezone(minutes)
    assert str(shim_zone) == repr(shim_zone)


@hypothesis.given(key=valid_zone_strategy)
def test_timezone_repr(key):
    zone = pds.timezone(key)

    assert key in repr(zone)


@hypothesis.given(dt=dt_strategy, key=valid_zone_strategy)
def test_localize_default_is_dst(dt, key):
    """Tests localize when is_dst is not specified.

    With pytz, is_dst defaults to False, but with the shims, if is_dst is
    unspecified we will assume that the user is indifferent about what to do
    and do whatever the underlying behavior of the time zone we're shimming
    around does.
    """

    shim_zone = pds.timezone(key)

    with pytest.warns(pds.PytzUsageWarning):
        dt_localized = shim_zone.localize(dt)

    dt_replaced = dt.replace(tzinfo=shim_zone)
    assert_dt_equivalent(dt_localized, dt_replaced)


@hypothesis.given(
    dt=dt_strategy,
    delta=hst.timedeltas(
        min_value=timedelta(days=-730), max_value=timedelta(days=730)
    ),
    key=valid_zone_strategy,
)
def test_normalize_pytz_zone(dt, delta, key):
    """Test that pds.normalize works on pytz-zoned datetimes.

    This allows you to take a zone you've normalized via pytz and convert
    it to the shim zone.
    """
    pytz_zone = pytz.timezone(key)
    shim_zone = pds.timezone(key)

    dt_pytz = pytz_zone.localize(dt) + delta
    dt_pytz_normalized = pytz_zone.normalize(dt_pytz)

    with pytest.warns(pds.PytzUsageWarning):
        dt_shim = shim_zone.normalize(dt_pytz)

    rounded_shim = round_normalized(dt_shim)
    rounded_pytz = round_normalized(dt_pytz_normalized)

    rounded_shim_naive = rounded_shim.replace(tzinfo=None)
    rounded_pytz_naive = rounded_pytz.replace(tzinfo=None)

    hypothesis.assume(MIN_DATETIME <= rounded_shim_naive <= MAX_DATETIME)
    assert rounded_shim_naive == rounded_pytz_naive


@pytest.mark.parametrize(
    "key, dt, offset", _zoneinfo_data.get_unambiguous_cases()
)
def test_localize_unambiguous_build_tzinfo(key, dt, offset):
    zone = pds.build_tzinfo(key, _zoneinfo_data.get_zone_file_obj(key))

    with pytest.warns(pds.PytzUsageWarning):
        dt_localized = zone.localize(dt)

    assert_dt_offset(dt_localized, offset)


def _skip_transition(key, zt):
    if not PY2:
        return False

    if key == "Europe/Dublin" and (
        zt.transition > datetime(1968, 1, 1)
        or zt.transition < datetime(1917, 1, 1)
    ):
        return True

    if key == "Africa/Casablanca" and zt.transition > datetime(2019, 1, 1):
        return True

    if key == "America/Santiago" and zt.transition < datetime(1912, 1, 1):
        return True

    return False


def _localize_fold_cases():
    cases = []

    def _add_case(key, dt, fold, offset, skip=False):
        case = (key, dt, fold, offset)

        if skip:
            case = pytest.param(*case, marks=pytest.mark.skip)

        cases.append(case)

    for key, zt in _zoneinfo_data.get_fold_cases():
        dt = zt.anomaly_start
        skip = _skip_transition(key, zt)
        _add_case(key, dt, 0, zt.offset_before, skip=skip)
        _add_case(key, dt, 1, zt.offset_after, skip=skip)

        dt = zt.anomaly_end - ONE_SECOND
        _add_case(key, dt, 0, zt.offset_before, skip=skip)
        _add_case(key, dt, 1, zt.offset_after, skip=skip)

    return tuple(cases)


@pytest.mark.parametrize("key, dt, fold, offset", _localize_fold_cases())
def test_localize_folds(key, dt, fold, offset):
    zone = pds.build_tzinfo(key, _zoneinfo_data.get_zone_file_obj(key))

    with pytest.warns(pds.PytzUsageWarning):
        dt_localized = zone.localize(enfold(dt, fold=fold))

    assert_dt_offset(dt_localized, offset)


def _localize_gap_cases():
    cases = []
    for key, zt in _zoneinfo_data.get_gap_cases():
        dt = zt.anomaly_start
        cases.append((key, dt, 0, zt.offset_before))
        cases.append((key, dt, 1, zt.offset_after))

        dt = zt.anomaly_end - ONE_SECOND
        cases.append((key, dt, 0, zt.offset_before))
        cases.append((key, dt, 1, zt.offset_after))

    return tuple(cases)


@pytest.mark.parametrize("key, dt, fold, offset", _localize_gap_cases())
@pytest.mark.skipif(
    PY2, reason="dateutil.tz doesn't properly support fold during gaps"
)
def test_localize_gap(key, dt, fold, offset):
    zone = pds.build_tzinfo(key, _zoneinfo_data.get_zone_file_obj(key))

    with pytest.warns(pds.PytzUsageWarning):
        dt_localized = zone.localize(enfold(dt, fold=fold))

    assert_dt_offset(dt_localized, offset)


def _folds_from_utc_cases():
    cases = []

    def _add_case(key, dt_utc, expected_fold, skip=False):
        case = (key, dt_utc, expected_fold)

        if skip:
            case = pytest.param(*case, marks=pytest.mark.skip)

        cases.append(case)

    for key, zt in _zoneinfo_data.get_fold_cases():
        dt_utc = zt.transition_utc
        dt_before_utc = dt_utc - ONE_SECOND
        dt_after_utc = dt_utc + ONE_SECOND

        skip = _skip_transition(key, zt)

        _add_case(key, dt_before_utc, 0, skip=skip)
        _add_case(key, dt_after_utc, 1, skip=skip)

    return tuple(cases)


@pytest.mark.parametrize("key, dt_utc, expected_fold", _folds_from_utc_cases())
def test_folds_from_utc(key, dt_utc, expected_fold):
    zone = pds.build_tzinfo(key, _zoneinfo_data.get_zone_file_obj(key))

    dt = dt_utc.astimezone(zone)
    assert get_fold(dt) == expected_fold


SHIM_ZONE_STRATEGY = hst.one_of(
    [
        valid_zone_strategy.map(pds.timezone),
        valid_zone_strategy.map(no_cache_timezone),
        offset_minute_strategy.map(pds.fixed_offset_timezone),
        hst.just(pds.UTC),
    ]
)


@hypothesis.given(shim_zone=SHIM_ZONE_STRATEGY)
def test_unwrap_shim(shim_zone):
    """Tests that .unwrap(_shim) always does the same as upgrade_tzinfo()."""
    assert shim_zone.unwrap_shim() is pds.helpers.upgrade_tzinfo(shim_zone)


@pytest.mark.skipif(
    PY2, reason="Currently dateutil does not have a .key attribute"
)
def test_wrap_zone_default_key():
    key = "America/New_York"
    zone = pds._compat.get_timezone(key)

    wrapped_zone = pds.wrap_zone(zone)

    assert str(wrapped_zone) == key


def test_wrap_zone_no_key():
    class TzInfo(tzinfo):
        def __init__(self):
            pass

    zone = TzInfo()

    with pytest.raises(TypeError):
        pds.wrap_zone(zone)


@hypothesis.given(shim_zone=SHIM_ZONE_STRATEGY)
def test_wrap_zone_same_object(shim_zone):
    """Tests unwrapping and rewrapping a shim zone.

    This should return the original shim object.
    """
    unwrapped = shim_zone.unwrap_shim()

    with pytest.warns(pds.PytzUsageWarning):
        key = shim_zone.zone

    rewrapped = pds.wrap_zone(unwrapped, key=key)

    assert rewrapped is shim_zone


@hypothesis.given(shim_zone=SHIM_ZONE_STRATEGY, key=ARBITRARY_KEY_STRATEGY)
def test_wrap_zone_new_key(shim_zone, key):
    """Test that wrap_zone can set arbitrary keys.

    There are a bunch of layers of caching here, so we want to make sure that
    if we wrap a zone that may already live in a cache with a new shim that has
    a different key, the new wrapper will reflect that, and the existing shim
    class won't be affected.
    """
    original_key = str(shim_zone)
    unwrapped = shim_zone.unwrap_shim()

    new_shim = pds.wrap_zone(unwrapped, key=key)

    assert str(new_shim) == key
    assert str(shim_zone) == original_key


@hypothesis.given(shim_zone=SHIM_ZONE_STRATEGY)
@pytest.mark.parametrize("copy_func", [copy.copy, copy.deepcopy,])
def test_copy(copy_func, shim_zone):
    shim_copy = copy_func(shim_zone)

    assert shim_copy is shim_zone


@hypothesis.given(shim_zone=SHIM_ZONE_STRATEGY, dt=dt_strategy)
@hypothesis.example(
    shim_zone=pds.timezone("America/New_York"), dt=datetime(2020, 11, 1, 1, 30)
)
@hypothesis.example(
    shim_zone=no_cache_timezone("America/New_York"),
    dt=datetime(2020, 11, 1, 1, 30),
)
@hypothesis.example(
    shim_zone=pds.timezone("America/New_York"),
    dt=enfold(datetime(2020, 11, 1, 1, 30), fold=1),
)
@conditional_examples(
    not PY2,
    examples=[
        hypothesis.example(
            shim_zone=no_cache_timezone("America/New_York"),
            dt=datetime(2020, 3, 8, 2, 30),
        ),
        hypothesis.example(
            shim_zone=no_cache_timezone("America/New_York"),
            dt=enfold(datetime(2020, 3, 8, 2, 30), fold=1),
        ),
    ],
)
def test_pickle_round_trip(shim_zone, dt):
    """Test that the results of a pickle round trip are identical to inputs.

    Ideally we would want some metric of equality on the pickled objects
    themselves, but with time zones object equality is usually equivalent to
    object identity, and that is not universally preserved in pickle round
    tripping on all Python versions and for all zones.
    """
    shim_copy = pickle.loads(pickle.dumps(shim_zone))

    dt = dt.replace(tzinfo=shim_zone)
    dt_rt = dt.replace(tzinfo=shim_copy)

    # PEP 495 says that the result of an inter-zone comparison between
    # datetimes where the offset depends on the fold is always False.
    if shim_zone is not shim_copy and dt != dt_rt:
        assert dt.dst() == dt_rt.dst()
        assert dt.utcoffset() == dt_rt.utcoffset()
        assert not datetime_unambiguous(dt)
        assert not datetime_unambiguous(dt_rt)
        return

    assert_dt_equivalent(dt, dt_rt)


@hypothesis.given(key=valid_zone_strategy)
@pytest.mark.skipif(PY2, reason="Singleton behavior is different in Python 2")
def test_pickle_returns_same_object(key):
    shim_zone = pds.timezone(key)

    shim_copy = pickle.loads(pickle.dumps(shim_zone))

    assert shim_zone is shim_copy


def test_utc_alias():
    assert pds.utc is pds.UTC



# -*- coding: utf-8 -*-
from datetime import datetime, timedelta

import hypothesis
import pytest
import pytz
from hypothesis import strategies as hst

import pytz_deprecation_shim as pds
from pytz_deprecation_shim import PytzUsageWarning

from ._common import (
    MAX_DATETIME,
    MIN_DATETIME,
    PY2,
    UTC,
    ZERO,
    assert_dt_equivalent,
    conditional_examples,
    dt_strategy,
    enfold,
    get_fold,
    offset_minute_strategy,
    round_timedelta,
    valid_zone_strategy,
)

_ARGENTINA_CITIES = [
    "Buenos_Aires",
    "Catamarca",
    "ComodRivadavia",
    "Cordoba",
    "Jujuy",
    "La_Rioja",
    "Mendoza",
    "Rio_Gallegos",
    "Rosario",
    "Salta",
    "San_Juan",
    "San_Luis",
    "Tucuman",
    "Ushuaia",
]

_ARGENTINA_ZONES = {"America/%s" % city for city in _ARGENTINA_CITIES}
_ARGENTINA_ZONES |= {
    "America/Argentina/%s" % city for city in _ARGENTINA_CITIES
}


@hypothesis.given(
    dt=dt_strategy, key=valid_zone_strategy, is_dst=hst.booleans()
)
@hypothesis.example(
    dt=enfold(datetime(2010, 11, 7, 1, 30), fold=1),
    key="America/New_York",
    is_dst=True,
)
@hypothesis.example(
    dt=enfold(datetime(2010, 11, 7, 1, 30), fold=1),
    key="America/New_York",
    is_dst=False,
)
@hypothesis.example(
    dt=datetime(2010, 11, 7, 1, 30), key="America/New_York", is_dst=True
)
@hypothesis.example(
    dt=datetime(2010, 11, 7, 1, 30), key="America/New_York", is_dst=False
)
@conditional_examples(
    not PY2,
    [
        hypothesis.example(
            dt=datetime(2009, 3, 29, 2), key="Europe/Amsterdam", is_dst=True
        ),
        hypothesis.example(
            dt=enfold(datetime(1933, 1, 1), fold=0),
            key="Asia/Kuching",
            is_dst=True,
        ),
        hypothesis.example(
            dt=enfold(datetime(1933, 1, 1), fold=1),
            key="Asia/Kuching",
            is_dst=True,
        ),
        hypothesis.example(
            dt=enfold(datetime(1933, 1, 1), fold=0),
            key="Asia/Kuching",
            is_dst=False,
        ),
        hypothesis.example(
            dt=enfold(datetime(1933, 1, 1), fold=1),
            key="Asia/Kuching",
            is_dst=False,
        ),
    ],
)
def test_localize_explicit_is_dst(dt, key, is_dst):
    pytz_zone = pytz.timezone(key)
    shim_zone = pds.timezone(key)

    dt_pytz = pytz_zone.localize(dt, is_dst=is_dst)
    with pytest.warns(PytzUsageWarning):
        dt_shim = shim_zone.localize(dt, is_dst=is_dst)

    assume_no_dst_inconsistency_bug(dt, key)
    assert_dt_equivalent(dt_shim, dt_pytz)


@hypothesis.given(dt=dt_strategy, key=valid_zone_strategy)
@hypothesis.example(dt=datetime(2018, 3, 25, 1, 30), key="Europe/London")
@hypothesis.example(dt=datetime(2018, 10, 28, 1, 30), key="Europe/London")
@hypothesis.example(dt=datetime(2004, 4, 4, 2, 30), key="America/New_York")
@hypothesis.example(dt=datetime(2004, 10, 31, 1, 30), key="America/New_York")
@conditional_examples(
    not PY2,
    examples=[
        hypothesis.example(
            dt=datetime(2018, 3, 25, 1, 30), key="Europe/Dublin"
        ),
        hypothesis.example(
            dt=datetime(2018, 10, 28, 1, 30), key="Europe/Dublin"
        ),
    ],
)
def test_localize_is_dst_none(dt, key):
    pytz_zone = pytz.timezone(key)
    shim_zone = pds.timezone(key)

    pytz_exc = shim_exc = dt_pytz = dt_shim = None
    try:
        dt_pytz = pytz_zone.localize(dt, is_dst=None)
    except pytz.InvalidTimeError as e:
        pytz_exc = e

    with pytest.warns(PytzUsageWarning):
        try:
            dt_shim = shim_zone.localize(dt, is_dst=None)
        except pds.InvalidTimeError as e:
            shim_exc = e

    # This section is triggered rarely. It currently occurs in Python 3 when
    # key = "Africa/Abidjan" and dt is 1912-01-01.
    if (dt_pytz is None) != (dt_shim is None):  # pragma: nocover
        uz = pds._compat.get_timezone(key)

        utc_off = uz.utcoffset(dt)
        hypothesis.assume(utc_off == round_timedelta(utc_off))

        utc_off_folded = uz.utcoffset(enfold(dt, fold=not get_fold(dt)))
        hypothesis.assume(utc_off_folded == round_timedelta(utc_off_folded))

    if dt_pytz:
        assume_no_dst_inconsistency_bug(dt, key, is_dst=False)
        assert_dt_equivalent(dt_pytz, dt_shim)

    if pytz_exc:
        assert repr(shim_exc) == repr(pytz_exc)
        assert str(shim_exc) == str(pytz_exc)
        assert isinstance(shim_exc, type(pytz_exc))


@hypothesis.given(
    dt=dt_strategy,
    delta=hst.timedeltas(
        min_value=timedelta(days=-730), max_value=timedelta(days=730)
    ),
    key=valid_zone_strategy,
)
def test_normalize_same_zone(dt, delta, key):
    """Test normalization after arithmetic.

    NOTE: There is actually a difference in semantics here, because with pytz
    zones, adding a timedelta and normalizing gives you absolute time
    arithmetic, not wall-time arithmetic. To emulate this, we do the addition
    of the shimmed zone in UTC.
    """
    pytz_zone = pytz.timezone(key)
    shim_zone = pds.timezone(key)

    dt_pytz = pytz_zone.localize(dt, is_dst=not get_fold(dt))
    dt_shim = dt.replace(tzinfo=shim_zone)

    hypothesis.assume(dt_pytz == dt_shim)

    dt_pytz_after_nn = dt_pytz + delta
    dt_shim_after_nn = (dt_shim.astimezone(UTC) + delta).astimezone(shim_zone)

    hypothesis.assume(
        MIN_DATETIME < dt_pytz_after_nn.replace(tzinfo=None) < MAX_DATETIME
    )

    dt_pytz_after = pytz_zone.normalize(dt_pytz_after_nn)

    with pytest.warns(PytzUsageWarning):
        dt_shim_after = shim_zone.normalize(dt_shim_after_nn)

    assert dt_shim_after_nn is dt_shim_after
    assume_no_dst_inconsistency_bug(dt + delta, key)
    assert_dt_equivalent(dt_pytz_after, dt_shim_after, round_dates=True)


@hypothesis.given(dt=dt_strategy, offset=offset_minute_strategy)
@hypothesis.example(dt=datetime(2020, 1, 1), offset=0)
@hypothesis.example(dt=datetime(2020, 2, 29), offset=0)
@hypothesis.example(dt=datetime(2020, 2, 29), offset=-1439)
@hypothesis.example(dt=datetime(2020, 2, 29), offset=1439)
def test_localize_fixed_offset(dt, offset):
    pytz_zone = pytz.FixedOffset(offset)
    shim_zone = pds.fixed_offset_timezone(offset)

    dt_pytz = pytz_zone.localize(dt)
    with pytest.warns(pds.PytzUsageWarning):
        dt_shim = shim_zone.localize(dt)

    assert dt_pytz == dt_shim
    assert dt_pytz.utcoffset() == dt_shim.utcoffset()


@hypothesis.given(key=valid_zone_strategy)
def test_zone_attribute(key):
    pytz_zone = pytz.timezone(key)
    shim_zone = pds.timezone(key)

    pytz_zone_value = pytz_zone.zone
    with pytest.warns(PytzUsageWarning):
        shim_zone_value = shim_zone.zone

    assert pytz_zone_value == shim_zone_value


@hypothesis.given(key=valid_zone_strategy)
def test_str(key):
    pytz_zone = pytz.timezone(key)
    shim_zone = pds.timezone(key)

    assert str(pytz_zone) == str(shim_zone)


def assume_no_dst_inconsistency_bug(dt, key, is_dst=False):  # prama: nocover
    # pytz and zoneinfo have bugs around the correct value for dst(), see, e.g.
    # Until those are fixed, we'll try to avoid these "sore spots" with a
    # combination of one-offs and rough heuristics.

    if len(key) == 3 and key.lower() == "utc":
        uz = pds._compat.UTC
    else:
        uz = pds._compat.get_timezone(key)

    ###########
    # One-offs
    if PY2:
        # https://github.com/dateutil/dateutil/issues/1048
        hypothesis.assume(
            uz.dst(dt)
            or not ((uz.dst(dt + timedelta(hours=24)) or ZERO) < ZERO)
        )

        # https://github.com/dateutil/dateutil/issues/1049
        hypothesis.assume(
            not (
                key == "Asia/Colombo"
                and datetime(1942, 1, 5) <= dt <= datetime(1945, 10, 17)
            )
        )

        # https://github.com/dateutil/dateutil/issues/1050
        hypothesis.assume(
            not (
                key == "America/Iqaluit"
                and datetime(1942, 7, 31) <= dt <= datetime(1945, 10, 1)
            )
        )

        # Possibly another manifestation of dateutil/dateutil#1050
        hypothesis.assume(
            not (
                (key == "MET" or key == "CET")
                and datetime(1916, 5, 1) <= dt <= datetime(1916, 10, 2)
            )
        )

        # dateutil isn't currently up to PEP 495's spec during ambiguous times,
        # which means occasionally it's an ambiguous time and neither side is
        # DST.
        from dateutil import tz

        hypothesis.assume(tz.datetime_exists(dt, uz))

    # bpo-40930: https://bugs.python.org/issue40930
    hypothesis.assume(
        not (
            key == "Pacific/Rarotonga"
            and datetime(1978, 11, 11) <= dt <= datetime(1991, 3, 2)
            and uz.dst(dt)
        )
    )
    hypothesis.assume(
        not (
            key == "America/Montevideo"
            and (
                datetime(1923, 1, 1) <= dt <= datetime(1927, 1, 1)
                or datetime(1942, 12, 14) <= dt <= datetime(1943, 3, 13)
            )
            and uz.dst(dt)
        )
    )

    # Argentina switched from -03 (STD) to -03 (DST) to -03 (STD) during this
    # interval, for whatever reason. pytz calls this dst() == 0, zoneinfo calls
    # this dst() == 1:00.
    hypothesis.assume(
        not (
            key in _ARGENTINA_ZONES
            and datetime(1999, 10, 3) <= dt <= datetime(2000, 3, 4)
        )
    )

    # bpo-40933: https://bugs.python.org/issue40933
    hypothesis.assume(
        not (
            key == "Europe/Minsk"
            and datetime(1941, 6, 27) <= dt <= datetime(1943, 3, 30)
        )
    )

    # Issue with pytz: America/Louisville transitioned from EST→CDT→EST in
    # 1974, but `pytz` returns timedelta(0) for CDT.
    hypothesis.assume(
        not (
            (
                key == "America/Louisville"
                or key == "America/Kentucky/Louisville"
                or key == "America/Indiana/Marengo"
            )
            and datetime(1974, 1, 6) <= dt <= datetime(1974, 10, 28)
        )
    )

    # Same deal with the RussiaAsia rule in 1991, a transition to DST with no
    # corresponding change in the offset, then a transition bac
    hypothesis.assume(
        not (
            key == "Asia/Qyzylorda"
            and datetime(1991, 3, 31) <= dt <= datetime(1991, 9, 30)
        )
    )

    # Issue with pytz: Europe/Paris went from CEST (+2, STD) → CEST (+2, DST) →
    # WEMT (+2, DST) → WEST (+1, DST) → WEMT (+2, DST) → CET (+1, STD) between
    # 3 April 1944 and 16 September 1945. pytz doesn't detect that WEMT is a
    # DST zone, though.
    hypothesis.assume(
        not (
            key == "Europe/Paris"
            and datetime(1944, 10, 8) <= dt <= datetime(1945, 4, 3)
        )
    )

    ###########
    # Bugs around incorrect double DST
    # https://github.com/stub42/pytz/issues/44
    # bpo-40931: https://bugs.python.org/issue40931
    pz = pytz.timezone(key)
    dt_pytz = pz.localize(dt, is_dst=is_dst)
    dt_uz = dt.replace(tzinfo=uz)
    dt_uz = enfold(dt_uz, fold=not is_dst)
    if abs(dt_pytz.dst()) <= timedelta(hours=1) and abs(
        dt_uz.dst() or timedelta(0)
    ) <= timedelta(hours=1):
        return

    hypothesis.assume(dt_uz.dst() == dt_pytz.dst())



# -*- coding: utf-8 -*-
import sys
from datetime import datetime, timedelta

import pytz
from hypothesis import strategies as hst

import pytz_deprecation_shim as pds

# Opportunistically bring in lru_cache, with a no-op if it's unavailable
# available
try:
    from functools import lru_cache
except ImportError:

    def lru_cache(maxsize):
        del maxsize

        def _(f):
            return f

        return _


PY2 = sys.version_info[0] == 2
ZERO = timedelta(0)
UTC = pds._compat.UTC

VALID_ZONES = sorted(pytz.all_timezones)
VALID_ZONE_SET = set(VALID_ZONES)

# There will be known inconsistencies between pytz and the other libraries
# right around the EPOCHALYPSE, because V1 files use 32-bit integers, and
# pytz does not support V2+ files.
NEGATIVE_EPOCHALYPSE = datetime(1901, 12, 13, 15, 45, 52)
EPOCHALYPSE = datetime(2038, 1, 18, 22, 14, 8)

MIN_DATETIME = NEGATIVE_EPOCHALYPSE + timedelta(days=365)
MAX_DATETIME = EPOCHALYPSE - timedelta(days=365)

MAX_OFFSET_MINUTES = 24 * 60 - 1  # pytz's range is (-1 day, 1 day)

valid_zone_strategy = hst.sampled_from(VALID_ZONES)
invalid_zone_strategy = hst.text().filter(lambda t: t not in VALID_ZONE_SET)
dt_strategy = hst.datetimes(min_value=MIN_DATETIME, max_value=MAX_DATETIME)
offset_minute_strategy = hst.integers(
    min_value=-MAX_OFFSET_MINUTES, max_value=MAX_OFFSET_MINUTES
)
invalid_offset_minute_strategy = hst.one_of(
    [
        hst.integers(max_value=-(MAX_OFFSET_MINUTES + 1)),
        hst.integers(min_value=(MAX_OFFSET_MINUTES + 1)),
    ]
)

# Helper functions
enfold = pds._compat.enfold
get_fold = pds._compat.get_fold


def datetime_unambiguous(dt):
    return dt.utcoffset() == enfold(dt, fold=not get_fold(dt)).utcoffset()


@lru_cache(128)
def round_timedelta(td):
    """Truncates a timedelta to the nearest minute."""
    if td == ZERO:
        return td

    tds = td.total_seconds()
    rounded = int((tds + 30) // 60) * 60
    return timedelta(seconds=rounded)


def round_normalized(dt):
    offset = dt.utcoffset()

    rounded_offset = round_timedelta(offset)
    if offset != rounded_offset:
        new_dt = dt + (rounded_offset - offset)
    else:
        new_dt = dt
    return enfold(new_dt, fold=pds._compat.get_fold(dt))


def assert_rounded_equal(actual, expected, **kwargs):
    actual_rounded = round_timedelta(actual, **kwargs)
    expected_rounded = round_timedelta(expected, **kwargs)

    assert actual_rounded == expected_rounded


def assert_dt_equivalent(actual, expected, round_dates=False):
    actual_utc_offset = actual.utcoffset()
    actual_dst = actual.dst()

    expected_utc_offset = expected.utcoffset()
    expected_dst = expected.dst()

    assert_rounded_equal(actual_utc_offset, expected_utc_offset)
    # There are too many inconsistencies and bugs in the calculation of dst()
    # in all three time zone libraries, so for the purposes of the shim, we
    # will be satisfied as long as the truthiness of the dst() calls is the
    # same between the two.
    # TODO: Uncomment this line when we've ironed out the bugs:
    # assert_rounded_equal(actual_dst, expected_dst)
    assert bool(actual_dst) == bool(expected_dst)

    assert actual.tzname() == expected.tzname()
    if round_dates:
        actual_naive = round_normalized(actual).replace(tzinfo=None)
        expected_naive = round_normalized(expected).replace(tzinfo=None)
    else:
        actual_naive = actual.replace(tzinfo=None)
        expected_naive = expected.replace(tzinfo=None)

    assert actual_naive == expected_naive


def assert_dt_offset(dt, offset):
    assert dt.tzname() == offset.tzname
    assert dt.utcoffset() == offset.utcoffset
    assert dt.dst() == offset.dst


def conditional_examples(cond, examples):
    if not cond:

        def _(f):
            return f

    else:

        def _(f):
            f_out = f
            for example in examples:
                f_out = example(f_out)
            return f_out

    return _
